/*

So far the mobs are behaving fairly intelligently, but if you've
played around with the team settings you may have noticed a problem:
mobs never change their targets.

Because these examples are simple mobs have no reason to prefer one
target over another, so there's no reason a mob would switch.
In a more complicated game, mobs might prefer a certain type of
target.

For example, red mobs might think of themselves as excellent fighters.
To show off their fighting skills they might choose to attack the most
powerful enemy. Blue mobs are weaker so they might choose to attack the
enemy with the lowest health.

There are a couple of changes we need to make here. Where we try to set
a mob's target (ex: in blue's call_for_help) we only change the mob's
target if they have no target. We introduce a new proc, change_target.
change_target takes a mob as a parameter and the mob (src) decides if
it should keep its current target or accept the new one.

This change is simple, but because of how the code was previously
organized we had to make changes in quite a few places.

Because we are now using the change_target proc, red and blue mobs can
both define their own behavior for changing targets.


  -- When you run this example --

In this example blue mobs are now on your team. Try to avoid the red mob
until it attacks a blue mob. If you move near the red mob it will switch
to you because you are more powerful than the blue mob.

*/

var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src)

	Stat()
		stat(src)
		stat("team", team)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0

	proc
		ai()
		hit_by(mob/m)
		attack(mob/target)
			show_hit(target)
			target.hit_by(src)

	Login()
		. = ..()

		power = 20
		health = 50

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			// We always want to look for a target, even if we already have
			// one. We might find a target that is more desirable than our
			// current one and switch to it.
			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
					sleep(10)
			else
				step_rand(src)
				sleep(20)

			spawn() ai()

		hit_by(mob/m)
			// We can use the new change_target proc to decide if
			// we should change to target the mob that attacked us.
			change_target(m)

		proc
			get_target()

			call_for_help()

			// The change_target proc takes mob/m as an argument and asks
			// the AI mob: do you want to change your target to m? To accept
			// this target as our new target we set target = m. If we return
			// from this proc before setting target, the mob is saying, "no,
			// I don't want to take m as my target."

			// This is the default behavior for change_target, the red and blue
			// mobs define their own behavior. This default behavior matches our
			// old behavior, mobs only take the new target if they didn't already
			// have a target
			change_target(mob/m)
				// If m is null we don't want to accept it. It was probably
				// an accident that null was passed to change_target.
				if(!m) return

				// We don't want the mobs to ever target a mob on their team.
				if(m.team == team) return

				// If we already have a target, ignore the new target
				if(target) return

				target = m

		red
			power = 30
			health = 40
			team = TEAM_2
			icon_state = "red"

			// Because we are now prioritizing targets we don't want
			// to target the first non-teammate within 2 tiles. A simple
			// way to use the new change_target proc is to have the red
			// mob try to change its target to every mob within two tiles.
			// If the new target is more desirable, it will switch. In the
			// end the mob will be targeting its most desirable target.
			get_target()
				for(var/mob/m in oview(2, src))
					change_target(m)

			change_target(mob/m)
				if(!m) return

				// We still want to ignore teammates
				if(m.team == team) return

				// If we didn't have a target, take the new target
				if(!target)
					target = m
					return

				// Only take the new target if it's more powerful
				// than our old target
				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 30

			team = TEAM_1
			icon_state = "blue"

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			change_target(mob/m)
				if(!m) return

				// We still want to ignore teammates
				if(m.team == team) return

				// If we didn't have a target, take the new target
				if(!target)
					target = m
					return

				// Only take the new target if it has less health
				// than our old target
				if(m.health < target.health)
					target = m