/*

Now we use a state machine to modify the behavior of the blue mobs.
Blue mobs can be in one of three states: HAPPY, FIGHT, or RUN_AWAY.
Their behavior is changed as follows:

    If the blue mob is happy and its health is below 5:
      The mob decides if it should stay and fight or run away (50-50 chance)
    If the blue mob is happy and its health is >= 5:
      The mob behaves as it always has (attack a target if it has one, otherwise wander)

    If the blue mob is in the fight state:
      if it has a target, attack it or move towards it
      if it doesn't have a target, go back to the happy state

    if the blue mob is in the run_away state:
      if it has a target, run away from it
      if it doesn't have a target, go back to the happy state

Blue mobs start in the happy state. The state machine for blue mobs is completely
separate from the state machine for red mobs. Even though they can both be in the
HAPPY state, their behavior in this state (and the states they can transition to
from it) may be very different.

This may seem overly complex for such a simple behavior but all the states are necessary.
It is important to realize that when the mob's health drops below 5 we want it to decide
*once* if it's going to stay and fight or run away. If we instead did something like this:

mob/blue/ai()
  if(target)
    if(health < 5)
      if(prob(50))
        step_away(src, target)
      else
        // attack target

In this example the mob would flip the coin again every AI cycle. One cycle it might decide
to run away, the next time it might decide to stay and fight. Because we want the decision
to happen once we have to transition to a different state.


  -- When you run this example --

In this example all types of mobs are on different teams. Watch and see what blue mobs do
when they get weak. Either you can attack the blue mob yourself or you can let the red mob
beat one up. This will show the transitions from HAPPY to FIGHT or RUN_AWAY.

You can also test the transitions from FIGHT or RUN_AWAY back to HAPPY. To do this, get the red
mob to attack a blue one. Once the blue mob transitions to FIGHT or RUN_AWAY approach the red
mob. Because your power is higher than the blue mob's power the red mob will attack you. When you
kill the red mob the blue mob (that was either fighting or fleeing) will transition back to happy.


Note: In earlier examples, all comments in the code were about new
changes. Because the code is getting more complicated I started leaving
some old comments in there.

*/

var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		ANGRY = "angry"
		HAPPY = "happy"

		FIGHT = "fight"
		RUN_AWAY = "run away"

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("state", m.state)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0

	proc
		ai()
		hit_by(mob/m)
		attack(mob/target)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= 1
			if(target.health <= 0)
				target.die()
				killed(target)

		// When a mob kills another, this event is called (src is the mob
		// that got the kill, m is the mob they killed). The default action
		// for this event is to do nothing. In this example we define a new
		// behavior for red mobs when they get a kill.
		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

	Login()
		. = ..()

		power = 20
		health = 20

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target

			// We add the state variable for all children of /mob/ai. We use the state
			// variable differently for different mobs:
			//   a red mob's state can be HAPPY or ANGRY
			//   a blue mob's state can be HAPPY, FIGHT, or RUN_AWAY.
			state = HAPPY

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return

			get_target()

			if(target)
				call_for_help()
				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
					sleep(10)
			else
				get_target()
				if(target)
					sleep(5)
				else
					step_rand(src)
					sleep(20)

			spawn() ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()
			call_for_help()

			// The default behavior is to accept the new target
			// only if you didn't already have one.
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return
				if(target) return

				target = m

		red
			power = 30
			health = 40
			team = TEAM_3
			icon_state = "red"

			// Red mobs start as being angry
			state = ANGRY

			// When a red mob kills another mob it becomes happy.
			// Note: for the killed proc to be called the red mob
			// has to be the one that gets the kill. If another mob
			// kills the red mob's target, the red mob will stay angry.
			killed(mob/m)
				state = HAPPY

			get_target()
				// If we have a target and its dead, lose that target.
				if(target && target.dead)
					target = null

				// Red mobs are only aggressive when they are angry.
				if(state == ANGRY)
					for(var/mob/m in oview(2, src))
						change_target(m)

			// red mobs target the enemy with the highest power.
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return

				// ignore dead targets
				if(m.dead) return

				if(!target)
					target = m
					return

				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 10

			team = TEAM_1
			icon_state = "blue"

			ai()
				if(dead) return

				// If the blue mob is happy it behaves almost the same way
				// it always has. The difference is that now, when the mob
				// gets below 5 health it decides if it should run away or
				// stay and fight.
				if(state == HAPPY)

					get_target()

					if(target)
						if(health < 5)
							// Decide if the mob is going to stay and fight or
							// run away. Announce the decision to the world,
							// even though it should be clear from its behavior.
							if(prob(50))
								world << "[name] decides to stay and fight."
								state = FIGHT
							else
								world << "[name] decides to run away."
								state = RUN_AWAY
							return ai()
						call_for_help()

						if(get_dist(src, target) > 1)
							step_to(src, target)
							sleep(5)
						else
							attack(target)
							sleep(10)
					else
						step_rand(src)
						sleep(20)

				// If it is in the FIGHT state it attacks its target. If it has
				// no target it goes back to the HAPPY state.
				else if(state == FIGHT)
					get_target()

					if(target)
						call_for_help()

						if(get_dist(src, target) > 1)
							step_to(src, target)
							sleep(5)
						else
							attack(target)
							sleep(10)
					else
						state = HAPPY
						sleep(5)

				// If the mob is in the RUN_AWAY state it runs away from its
				// target and calls for help
				else if(state == RUN_AWAY)
					get_target()

					if(target)
						call_for_help()
						step_away(src, target)
						sleep(3)

					// If the mob no longer has a target (this would happen
					// if its target died) it goes back to the HAPPY state.
					else
						state = HAPPY
						sleep(5)

				spawn() ai()

			get_target()
				// blue mobs don't look for targets, but if they have a
				// target and it dies we want them to lose that target.
				if(target && target.dead)
					target = null

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// blue mobs target the enemy with the lowest health
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return

				// ignore dead targets
				if(m.dead) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m
