/*

The mob's ai proc is a loop. It doesn't use for or while, it creates
a loop by calling the function over again. Here is the ai proc from
the first example, code-01.dm:

	ai()
		step_rand(src)
		spawn(10) ai()

What does spawn(10) do and why is it necessary?

the spawn proc delays execution of code but lets the current proc
continue its execution. Here's an example:

	spawn(10)
		world << "A"
	world << "B"

The execution of world << "A" is delayed by one second, so this program
would output "B" then "A".

What would happen if we changed our ai proc to this:

	ai()
		step_rand(src)
		sleep(10)
		ai()

This appears to do the same thing. The ai() proc is called again one
second after the call to step_rand. This seems correct, but you'll run
into trouble. It'll take some explaining to understand why.

The call stack is a list of function calls. When you call a function,
the function is added to the front of the list. When that function ends
(when its end is reached or when you call return) that function is removed
from the list. The function that is now at the front of the list is what
we return to. For example:

	turf
		Click()
			a()

	proc
		a()
			world << "A"
			b()

		b()
			world << "B"

If we run this program and click on a turf, here's how the call stack
will change over time:

Action						Call Stack
user clicks on a turf		/turf/proc/Click
Click calls a				/turf/proc/Click, /proc/a
a calls b					/turf/proc/Click, /proc/a, /proc/a
b returns					/turf/proc/Click, /proc/a
a returns					/turf/proc/Click
Click returns				(empty)

Let's see how the call stack behaves in our last ai proc:

	ai()
		step_rand(src)
		sleep(10)
		ai()

ai is called from mob.New	/mob/ai/proc/New, /mob/ai/proc/ai
ai calls ai					/mob/ai/proc/New, /mob/ai/proc/ai, /mob/ai/proc/ai
ai calls ai					/mob/ai/proc/New, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai
ai calls ai					/mob/ai/proc/New, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai
ai calls ai					/mob/ai/proc/New, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai, /mob/ai/proc/ai

The ai proc never returns! Because of this, the call stack grows and
grows. The call stack is stored in memory, and because you have a
limited amount of memory you'll eventually run out.

Ok, that way of creating the AI loop is out. Let's look at another popular
method:

	ai()
		while(1)
			step_rand(src)
			sleep(10)

This looks better. The proc doesn't call itself so we shouldn't run
into the call stack problem -- or will we?

Let's take a look at the state machine from code-09a.dm. In particular,
look at the indicated line.

	ai()
		if(dead) return

		if(state == HAPPY)
			get_target()

			if(target)
				if(health < 5)
					// Decide if the mob is going to stay and fight or
					// run away. Announce the decision to the world,
					// even though it should be clear from its behavior.
					if(prob(50))
						world << "[name] decides to stay and fight."
						state = FIGHT
					else
						world << "[name] decides to run away."
						state = RUN_AWAY
Look at this line ---->	return ai()
				...
			else
				step_rand(src)
				sleep(20)
		else if(state == FIGHT) ...
		else if(state == RUN_AWAY) ...
		spawn() ai()

When the mob is in the HAPPY state and is below 5 health it decides
if it should transition to FIGHT or RUN_AWAY. Once it makes a transition
we call "return ai()". Calling return will exit this current instance of
the ai loop, but first it calls ai() again. Because the mob changed
states we want to execute another ai cycle immediately so that the mob
will perform an action that is appropriate for its new state.

This code will not have the call stack problem because every call to
the ai proc terminates. A call either reaches the end, spawns the next
call, then ends. Or, the call reaches a "return ai()" line, which returns
after the recursive call ends.

If we didn't want to use "return ai()", the alternative would be
something like this:

		if(state == HAPPY)
			get_target()

			if(target)
				if(health < 5)
					if(prob(50))
						world << "[name] decides to stay and fight."
						state = FIGHT
						// fight code
					else
						world << "[name] decides to run away."
						state = RUN_AWAY
						// run away code
				...
			else
				step_rand(src)
				sleep(20)
		else if(state == FIGHT)
			// fight code
		else if(state == RUN_AWAY)
			// run away code
		spawn() ai()

We'd have to repeat the code for fighting or running away within the
block that handles the HAPPY state. We'll go into more detail in code-10.dm,
but repeating code like this is a bad idea.

If we had used a while loop for our ai proc, like this:

	ai()
		while(1)
			if(dead) return
			if(state == HAPPY)
				...
					return ai()
				...
			else if(state == FIGHT)
				...
			else if(state == RUN_AWAY)
				...
			sleep(10)

We could still run into the call stack problem. When the mob transitions
to a new state and calls "return ai()" it adds a function to the call
stack and this new function will never end (so it will never be removed
from the call stack). It may not happen often enough that we'll run out
of memory, but it will still take up an unnecessary amount of memory.

Instead we can use something you may have never heard of: loop labels.

	ai()
		ai_loop:
			while(1)
				if(dead) return
				if(state == HAPPY)
					...
						continue ai_loop
					...
				else if(state == FIGHT)
					...
				else if(state == RUN_AWAY)
					...
				sleep(10)

The "ai_loop:" before the while loop is a loop label. To assign the
while loop to that label we indent the while loop. When inside a loop,
the continue command aborts the current iteration of the loop and
starts the next iteration. If you don't specify a label it will start
the next iteration of the innermost loop. We need to use a label
because our code might look like this someday:

	ai()
		ai_loop:
			while(1)
				if(dead) return
				if(state == HAPPY)
					...
						for(var/i = 1 to 10)
							...
							continue ai_loop
					...
				else if(state == FIGHT)
					...
				else if(state == RUN_AWAY)
					...
				sleep(10)

The continue statement is inside of two loops (the while and the for). Just
saying "continue" would start the next iteration of the for loop. In this
case we have to use a label.

Loop labels are strange and require you to indent your proc by another tab,
With the number of if statements you'll end up having inside your ai proc
you don't need more things contributing to indentation.

*/
