/*

This example is the same as code-09.dm, this example just shows some
ways to clean up the code. As the AI gets more complicated this can
be especially important.

Here is the code for /mob/ai/blue's ai proc in the last example:

				if(state == HAPPY)
					get_target()
					if(target)
						if(health < 5)
							// decide to fight or run away
				*		call_for_help()
				*		if(get_dist(src, target) > 1)
				*			step_to(src, target)
				*			sleep(5)
				*		else
				*			attack(target)
				*			sleep(10)
					else
						step_rand(src)
						sleep(20)

				else if(state == FIGHT)
					get_target()
					if(target)
				*		call_for_help()
				*		if(get_dist(src, target) > 1)
				*			step_to(src, target)
				*			sleep(5)
				*		else
				*			attack(target)
				*			sleep(10)
					else
						state = HAPPY
						sleep(5)

				else if(state == RUN_AWAY)
					get_target()

					if(target)
						call_for_help()
						step_away(src, target)
						sleep(3)
					else
						state = HAPPY
						sleep(5)

The sections marked with *s are identical. In these two situations the mob is in
different states but we want it to use the same behavior. Repeating the code in both
places can lead to problems. If we want to change this behavior we have to change it
in both places. That doesn't sound bad because there are only two places where we need
to make changes, but as the complexity grows and we add more states, we may repeat that
same code several more times.

To address this we create a new proc called pursue_target which does exactly what the
marked lines do. This way we can call pursue_target in each situation and get the same
behavior, while only having one place we need to change things if we want to modify the
behavior.

Another way to address this issue would be to combine states that are similar. For example,
since the HAPPY and FIGHT states are very similar, instead of the code looking like this:

    if(state == HAPPY)
        // actions for the happy state
    else if(state == FIGHT)
        // actions for the fight state
    else if(state == RUN_AWAY)
        // actions for the run_away state

The code would look like this:

    if(state == HAPPY || state == FIGHT)
        // actions for the happy and fight states
    else if(state == RUN_AWAY)
        // actions for the run_away state

It might not always be possible, but sometimes this approach might let you write cleaner,
more organized code and cut down on repeated code. There is an example of this approach
in the ai proc for /mob/ai/blue.


Note: In earlier examples, all comments in the code were about new
changes. Because the code is getting more complicated I started leaving
some old comments in there.

*/

var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		ANGRY = "angry"
		HAPPY = "happy"

		FIGHT = "fight"
		RUN_AWAY = "run away"

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("state", m.state)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0

	proc
		ai()
		hit_by(mob/m)
		attack(mob/target)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= 1
			if(target.health <= 0)
				target.die()
				killed(target)

		// When a mob kills another, this event is called (src is the mob
		// that got the kill, m is the mob they killed). The default action
		// for this event is to do nothing. In this example we define a new
		// behavior for red mobs when they get a kill.
		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

	Login()
		. = ..()

		power = 20
		health = 20

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target

			// We add the state variable for all children of /mob/ai, even
			// though we only use it for red mobs in this example.
			state = HAPPY

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return

			get_target()

			if(target)
				pursue_target()

				// The call to pursue_target replaces these seven lines
				/*
				call_for_help()
				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
				 	sleep(10)
				 */
			else
				get_target()
				if(target)
					sleep(5)
				else
					step_rand(src)
					sleep(20)

			spawn() ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()
			call_for_help()

			// The default behavior is to accept the new target
			// only if you didn't already have one.
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return
				if(target) return

				target = m


			// We add the pursue_target proc. This proc determines what
			// happens when the mob decides to pursue its target. Previously
			// we didn't need this proc, but if you look at the ai proc for
			// /mob/ai/blue in the previous example you can see this code get
			// repeated for the HAPPY and FIGHT states.
			pursue_target()
				call_for_help()
				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
					sleep(10)

		red
			power = 30
			health = 40
			team = TEAM_3
			icon_state = "red"

			// Red mobs start as being angry
			state = ANGRY

			// When a red mob kills another mob it becomes happy.
			// Note: for the killed proc to be called the red mob
			// has to be the one that gets the kill. If another mob
			// kills the red mob's target, the red mob will stay angry.
			killed(mob/m)
				state = HAPPY

			get_target()
				// If we have a target and its dead, lose that target.
				if(target && target.dead)
					target = null

				// Red mobs are only aggressive when they are angry.
				if(state == ANGRY)
					for(var/mob/m in oview(2, src))
						change_target(m)

			// red mobs target the enemy with the highest power.
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return

				// ignore dead targets
				if(m.dead) return

				if(!target)
					target = m
					return

				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 10

			team = TEAM_1
			icon_state = "blue"

			ai()
				if(dead) return

				// By using the pursue_target proc we can shorten this
				// code a little bit. Because different states may share
				// the same behaviors we can come up with ways to shorten
				if(state == HAPPY)
					get_target()

					if(target)
						if(health < 5)
							if(prob(50))
								world << "[name] decides to stay and fight."
								state = FIGHT
							else
								world << "[name] decides to run away."
								state = RUN_AWAY
							return ai()

						pursue_target()
					else
						step_rand(src)
						sleep(20)

				else if(state == FIGHT)
					get_target()

					if(target)
						pursue_target()
					else
						state = HAPPY
						sleep(5)

				else if(state == RUN_AWAY)
					get_target()

					if(target)
						call_for_help()
						step_away(src, target)
						sleep(3)
					else
						state = HAPPY
						sleep(5)

				// Having separate if statements for each state is a
				// good way to keep things organized. It makes it easy
				// to see what the behavior is for each state. If two
				// states have similar behavior you might want to do
				// something like this:
				/*
				// Because HAPPY and FIGHT have a lot of overlap, we'll
				// treat them as being the same and make distinctions
				// inside this if statement when we need to.
				if(state == HAPPY || state == FIGHT)
					get_target()

					if(target)
						// We can only make these transitions if we're
						// in the HAPPY state.
						if(state == HAPPY && health < 5)
							if(prob(50))
								world << "[name] decides to stay and fight."
								state = FIGHT
							else
								world << "[name] decides to run away."
								state = RUN_AWAY
								return ai()

						pursue_target()
					else
						state = HAPPY
						step_rand(src)
						sleep(20)

				else if(state == RUN_AWAY)
					get_target()

					if(target)
						call_for_help()
						step_away(src, target)
						sleep(3)
					else
						state = HAPPY
						sleep(5)
				*/

				spawn() ai()

			get_target()
				// blue mobs don't look for targets, but if they have a
				// target and it dies we want them to lose that target.
				if(target && target.dead)
					target = null

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// blue mobs target the enemy with the lowest health
			change_target(mob/m)
				if(!m) return
				if(m.team == team) return

				// ignore dead targets
				if(m.dead) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m
