/*

 *** Intermission, continued ***

Yes, yes, more text =)

So far our mobs have been able to do two things: move and attack. Even with
just these two actions we can create mobs that have personalities. Red mobs
are big, angry brutes that seek out the strongest opponent and never ask others
for help. Blue mobs are more passive and tend to run away when they get weak.

We've come a long way from random wandering, but there's still a long way to go.

Even the smartest mob we've made so far had a very limited decision-making
ability. To decide whether they should stay and fight or run away, blue mobs
just flip a coin! Once they've made their decision they act accordingly, but
the decision itself is not made intelligently. We would like to create mobs
that can assess the current situation, consider their options, and make an
intelligent decision.

We'll get to that eventually, but first we need to understand the concept of
utility.

  -- Utility --

Utility is a measure of how good a situation is from a particular mob's
point of view. If a mob has five allies that are still alive and their
only remaining enemy is low on health, we can say that this situation has
a high utility. It certainly has a higher utility than the situation where
all of the mob's allies are dead and his enemies are at full health.

We can use utility to decide what if one situation is better than another.
If we know what actions a mob can perform and how each action would affect
the current situation, we can evaluate the utility of each resulting situation
to figure out which action appears to be the best (in other words, the action
that causes the greatest increase in utility).


  -- Example --

At the bottom of this file we have an example of this. The example does not
build on any of the previous examples. We'll start with a very simple example
to see what utility does (and doesn't) do, then we'll build up more complex
examples.

In this example there is a single mob that you control. To move you don't use
the arrow keys, you click on a tile. The tile you click on becomes your destination.
Once we know your destination we can compute the utility of any turf as follows:

	utility(turf/t)
		if(t.density) return 0
		return 100 - get_dist(t, destination)

This should make sense. The utility function should return a higher number for
more desirable locations. We cannot move to a dense turf, so they have a utility
of zero. The utility of a non-dense turf is based on its distance to our
destination. The closer it is, the higher the utillity is (since we're taking
100 minus distance, the closer we are the lower the distance is, so the higher
100 - distance is).

Each AI cycle evaluates the utility of our current location and the utility
of all turfs we can move to (the 8 adjacent turfs). We find the tile with the
highest utility and move to that one.

If you haven't already, feel free to run the example now.

  -- Problem --

The result, as you might expect, is that the mob moves towards the tile you click
on. You might realize a problem. In a situation like the one pictured below, the
mob never reaches its destination:

     ########
            #
       mob  #     destination
            #
     ########

  # = wall

It should be clear why this happens. The only valid move the mob can make is
to the left. Since the tile to its left is further from the destination than
its current location it has a lower utility. The mob's current location is
the highest utility it can achieve so it stays put.

There are two ways you can think about the problem. Each way shows us an important
concept at work here:

1. Our utility function is flawed. The distance to the destination does not always
correspond to the number of moves it would take to get there. Because of this, in
the situation pictured above our utility function is not a good estimate of the true
distance to the destination.

This is because our utility function uses a heuristic. A heuristic is a simple rule
that you use to estimate a more complicated value.

In this case we use a heuristic because we don't know how many steps it would
actually take to get to the destination. To find the actual number of steps we'd
have to find the complete path, but if we knew the path we wouldn't need to use
a utility function to determine which way to move, we'd just use the path.

The other way to look at the problem is:

2. Our utility function is okay, we're just not using it properly. If we could
look ahead more than one step we could see that even though a move to the left
reduces our utility, it can lead to some moves that increase our utility even
more (by moving down, right, right, and up to the destination).

The problem is that our strategy picks the single action that gives us the greatest
utility, not a sequence of actions that yields the greatest utility. This type of
strategy is called a greedy strategy. The name makes sense: the mob is greedy because
it wants the most utility.


  -- Conclusion --

I'm sure you can find a much easier way to write the same program as the example
below (you could use step_towards), but the example is meant to show the concepts
of utility functions, heuristics, and greedy strategies. Even though these concepts
do not produce an ideal solution in this example, we will use these concepts more
intelligently in later examples to create better AI.

*/

turf
	Click()
		usr.destination = src
		usr.marker.loc = src

	proc
		neighbors()
			var/list/l = list()
			for(var/turf/t in oview(1, src))
				l += t
			return l

mob
	var
		turf/destination
		obj/destination_marker/marker
	Login()
		..()

		marker = new /obj/destination_marker(loc)

		icon_state = "green"
		destination = loc

		// We're going to use the same map we've been using
		// for all of the examples so far. It has red and blue
		// mobs placed on it, but we don't want them getting
		// in our way for this example.
		for(var/mob/m in world)
			if(m != src)
				del m

		spawn(5) ai()

	proc
		// This is our utility function, as it was explained above:
		// The utility function should return a higher number for
		// more desirable locations. We cannot move to a dense turf,
		// so they have a utility of zero. The utility of a non-dense
		// turf is based on its distance to our destination. The closer
		// it is, the higher the utillity is.
		utility(turf/t)
			if(t.density) return 0

			return 100 - get_dist(t, destination)

		ai()
			var/turf/current_loc = loc
			var/current_situation = utility(current_loc)

			var/best_utility = current_situation
			var/turf/best_move = current_loc

			// the set of possible moves we can make is the set of
			// tiles that are adjacent to our current location (plus
			// our current location, because we can stay still). We
			// find the move that gives the the highest utility and
			// execute that move.
			for(var/turf/t in current_loc.neighbors())
				if(utility(t) > best_utility)
					best_utility = utility(t)
					best_move = t

			Move(best_move)

			spawn(3) ai()

	// we don't use these, but we need to define them because
	// they're placed on the map. I don't want to have a different
	// map for every example.
	ai
		red
		blue

// our destination marker
obj/destination_marker/icon_state = "x"

// prevent keyboard input
client
	North()
	South()
	East()
	West()
	Northwest()
	Southwest()
	Northeast()
	Southeast()