/*

Now we're back to building on old examples. This example builds on code-08a.dm,
but changes the way attacking works.

We're going to take an old example and add some new behavior
based on heuristics, greedy strategies, and utility functions.

In this example we look at how we can deal with multiple types
of abilities. Blue mobs can either deal 1 damage to an enemy or
heal an ally to restore 10 health points. We use a utility function
to evaluate which is the better move.

How exactly should our utility function work?

Restoring 10 points of an ally's heatlh should not always give the
same increase in utility. If the ally is at 90 / 100 health it
should give less utility than if they were at 10 / 100 health. There
are two reasons for this:

  1. If we're going to heal an ally, we should heal the weakest one.
  2. If an ally has lots of health, we shouldn't waste time healing it.

These reasons should make sense. The utility functions are implemented
in the blue mob's get_target proc. The utility functions I used are:

  For attacking an enemy:
    utility = 10 / (enemy.health + 10)

  For healing an ally:
    utility = 5 / ally.health

Math may not be your strongest suit, but that's okay. Both values
are fractions with the mobs health in the denominator (the bottom).
As the denominator increases, the value of the fraction decreases.
This way we get the desired effect: healing a weaker ally gives a
larget increase in utility.

You can plug in numbers to see how things work out. For example:

    ally is at 10 health, enemy is at 15 health.
    utility for healing = 5 / 10 = 0.5
    utility for attacking = 10 / (15 + 10) = 0.4

Because the ally is weak the utility for healing him is higher.
If the ally at 15 health and the enemy also had 15 health:

    utility for healing = 5 / 15 = 0.333
    utility for attacking = 10 / (15 + 10) = 0.4

In this case the utility for attacking is higher, so the mob will
attack.


  -- When you run this example --

In this example attack a mob until it is weak and see if its blue allies
heal it. When a blue mob heals an ally its target will become that ally,
in addition to the visual effect of healing you can watch the statpanel to
see when a blue mob switches from targeting you to target one of its allies.

*/


var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src, 3)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0
		max_health = 0

	proc
		ai()
		hit_by(mob/m)
		healed_by(mob/m)

		// attacks now deal various amounts of damage.
		attack(mob/target, damage)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= damage
			if(target.health <= 0)
				target.die()
				killed(target)

		// mobs can now heal other mobs, so we need
		// this helper function.
		heal(mob/target, healing)
			if(target.dead) return

			show_heal(target)
			target.healed_by(src)

			target.health += healing
			if(target.health > target.max_health)
				target.health = target.max_health

		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

				if(client)
					client << "You will resurrect in 3 seconds..."
					sleep(30)
					loc = locate(2,2,1)
					health = max_health
					dead = 0
					icon_state = "green"

	Login()
		. = ..()

		power = 20
		health = 50
		max_health = 50

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return
			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target, 2)
					sleep(10)
			else
				step_rand(src)
				sleep(20)

			spawn() ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()

			call_for_help()

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return
				if(target) return
				target = m

		red
			power = 30
			health = 40
			max_health = 40
			team = TEAM_2
			icon_state = "red"

			get_target()
				for(var/mob/m in oview(2, src))
					change_target(m)

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 30
			max_health = 30

			team = TEAM_2
			icon_state = "blue"

			ai()
				if(dead) return
				get_target()

				if(target)
					// if the target is on your team, heal them
					if(target.team == team)
						heal(target, 10)

					// otherwise the target is an enemy and we attack them
					else
						call_for_help()
						attack(target, 1)
				else
					step_rand(src)
					sleep(20)

				spawn() ai()

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// This is where we implement and use the utility functions
			// to determine what action we take.
			get_target()
				target = null
				var/most_utility = 0

				// We can heal any ally within 5 tiles, so look
				// at each ally and estimate the utility we'd gain
				// by healing them.
				for(var/mob/m in view(5, src))
					if(!m.dead && m.team == team)
						if(m.health == m.max_health) continue
						var/utility = 5 / (m.health)
						if(utility > most_utility)
							most_utility = utility
							target = m

				// We can attack any enemy within 3 tiles, so look
				// at each enemy and estimate the utility we'd gain
				// by attacking them.
				for(var/mob/m in oview(3, src))
					if(!m.dead && m.team != team)
						var/utility = 10 / (m.health + 10)
						if(utility > most_utility)
							most_utility = utility
							target = m

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m