/*

The example in code-13.dm used a utility function to decide
if the blue mobs should heal an ally or attack an enemy.
What we neglected was movement.

In the previous example the blue mobs would only consider
actions that they could perform immediately. The range on
their heal ability was 5 tiles, so they only looked for weak
allies within 5 tiles. They might have a severely wounded
ally that is 6 tiles away, but the AI in that example would
never consider moving a tile to heal an ally.

In this example we use utility functions to determine what
action should be taken, then we use a state machine to carry
out that action.

The blue mob can be in one of three states: NORMAL, ATTACK,
or HEAL. Only in the NORMAL state will the mob use utility
functions to decide what action should be taken. If it decides
to heal an ally it transitions to the HEAL state. Once in
the HEAL state it moves towards the ally until within range,
heals the ally, then transitions back to the NORMAL state.

Similarly, for attacking the blue mob moves towards the enemy
until it is within range (three tiles), performs the attack,
then transitions back to the NORMAL state.


  -- When you run this example --

When you run this example attack the red mob and get it weak,
then run away from it. The red mob will follow because you are
its target. See if the blue mobs follow to heal it and continue
attacking you.


Exercise:

When the blue mob decides to perform an action it needs to get
within range of its target. There could be a lot of time between
deciding to perform an action and actually performing it,
especially if it has to chase after its target. In this time
the mob's situation could drastically change. If a drastic change
occurs, the best move might be to abort the current action and
perform a different one. How would you implement this?

Hint: You need to be careful that the mob doesn't keep changing
its mind, consider this situation:

    ALLY               BLUE MOB               ALLY

The blue mob is between two allies who keep taking damage. Both are
out of its healing range. As it moves towards one, the other takes
damage. You don't want the blue mob to run back and forth because
it keeps changing its mind and never get to cast a heal.

*/

var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		NORMAL = 1
		HEAL = 2
		ATTACK = 3

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src, 3)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0
		max_health = 0

	proc
		ai()
		hit_by(mob/m)
		healed_by(mob/m)

		attack(mob/target, damage)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= damage
			if(target.health <= 0)
				target.die()
				killed(target)

		heal(mob/target, healing)
			if(target.dead) return

			show_heal(target)
			target.healed_by(src)

			target.health += healing
			if(target.health > target.max_health)
				target.health = target.max_health

		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

				if(client)
					client << "You will resurrect in 3 seconds..."
					sleep(30)
					loc = locate(2,2,1)
					health = max_health
					dead = 0
					icon_state = "green"

	Login()
		. = ..()

		power = 20
		health = 50
		max_health = 50

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target
			state

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return
			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target, 2)
					sleep(10)
			else
				step_rand(src)
				sleep(20)

			spawn() ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()

			call_for_help()

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return
				if(target) return
				target = m

		red
			power = 30
			health = 40
			max_health = 40
			team = TEAM_2
			icon_state = "red"

			get_target()
				for(var/mob/m in oview(2, src))
					change_target(m)

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 30
			max_health = 30

			team = TEAM_2
			icon_state = "blue"

			state = NORMAL

			// The code for the blue mob's ai proc is a bit longer now
			// because we've added the state machine, but its very
			// straightforward.
			ai()
				if(dead) return

				if(state == NORMAL)
					// get_target uses the utility functions to decide what action
					// should be taken, so we only call if if we're in the NORMAL state.
					// If we're in the HEAL or ATTACK states we are in the process of
					// executing an action and we don't want to switch actions.
					get_target()

					// If we get a target we transition to the ATTACK or HEAL state.
					if(target)
						if(target.team == team)
							state = HEAL
							sleep(1)
							return ai()
						else
							state = ATTACK
							sleep(1)
							return ai()
					else
						step_rand(src)
						sleep(20)

				else if(state == ATTACK)
					// If we've lost our target, go back to the NORMAL state
					if(!target || target.dead)
						state = NORMAL
						sleep(1)
						return ai()

					// If we're within range, make the attack and transition
					// back to the NORMAL state.
					if(get_dist(src, target) <= 3)
						attack(target, 2)
						sleep(10)
						state = NORMAL

					// otherwise move closer.
					else
						step_to(src, target)
						sleep(5)

				// Healing is very similar to attacking. If you're not within
				// range you need to move closer. Once you're within range,
				// perform the heal and go back to the NORMAL state.
				else if(state == HEAL)
					// If we've lost our target, go back to the NORMAL state
					if(!target || target.dead)
						state = NORMAL
						sleep(1)
						return ai()

					if(get_dist(src, target) <= 5)
						heal(target, 10)
						sleep(10)
						state = NORMAL
					else
						step_to(src, target)
						sleep(5)

				spawn() ai()

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// The only changes made to this function were to
			// increase the range that the mob considers. Even though
			// the blue mob can only heal allies within 5 tiles, it
			// looks within 8 tiles to find a weak ally. A similar
			// change was made for finding an enemy to attack.
			get_target()
				target = null
				var/most_utility = 0

				for(var/mob/m in view(8, src))
					if(!m.dead && m.team == team)
						if(m.health == m.max_health) continue
						var/utility = 5 / (m.health)
						if(utility > most_utility)
							most_utility = utility
							target = m

				for(var/mob/m in oview(8, src))
					if(!m.dead && m.team != team)
						var/utility = 10 / (m.health + 10)
						if(utility > most_utility)
							most_utility = utility
							target = m

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m