/*

 ** IMPORTANT NOTE ** use map-2.dmi for this example.

To switch maps, in the file tab on the left uncheck map-1.dmm and
check off map-2.dmm.


Some behaviors would be very hard to model, harder than its worth.

How would you make a mob that runs and hides when it gets weak?

We already saw an example where blue mobs had a chance to run away when
they got weak, but running away is different from hiding. We want the
mobs to find a spot that is out of the way, somehow disguised, or hard
to find, then move to that location and sit there.

We could come up with a heuristic that looks at every tile on the map
and determines how good of a hiding spot it would be. This could be
very time consuming and still might not produce ideal results.

For example, one heuristic might say that a good hiding spot is a tile
that has lots of dense tiles next to it, like a dead end of a narrow
hallway. This heuristic isn't horrible: it gives a low score to open
areas which is a good thing. But consider this map and the points
marked a and b:

         ###
  ########a########
  #               #    # = wall
  #               #
  #               #
  # #             #
  # #         #####
  # #         #   #
  # ########### b #
  #               #
  #################

Point b has no dense turfs immediately next to it, but would seem to a
human as a a better hiding spot than point a. Point a has five dense
tiles next to it, making it the best hiding spot on the map, but we can
tell that point a, though it is against the wall, is in a fairly open,
easy to reach spot and would make a bad hiding spot.

We could develop a more complex heuristic. For each turf we might pick
a few random points on the map and count the number of steps it would
take to walk from those random points to the turf. A hiding spot would
be a turf that is very distant from the random points. Not only would
this heuristic have problems with correctness, but computing these values
for every turf on the map would be *very* time consuming.

Instead of doing this, we cheat!

Yes, it's that easy. If you haven't already, take a look at map-2.dmm.
At coordinates 6,5 and 13,11 are objs called hiding spots. These markers
turn invisible when the game starts up, but AI-controlled players can use
these objects to know where good hiding spots are.


  -- When you run this example --

When you run this example, hit a blue mob enough times that it drops
below ten health and watch it run away. Try to keep up with it!

*/

var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		NORMAL = 1
		HIDE = 2

turf
	wall
		opacity = 1

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src, 3)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0
		max_health = 0

	proc
		ai()
		hit_by(mob/m)
		healed_by(mob/m)

		attack(mob/target, damage)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= damage
			if(target.health <= 0)
				target.die()
				killed(target)

		heal(mob/target, healing)
			if(target.dead) return

			show_heal(target)
			target.healed_by(src)

			target.health += healing
			if(target.health > target.max_health)
				target.health = target.max_health

		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

				if(client)
					client << "You will resurrect in 3 seconds..."
					sleep(30)
					loc = locate(2,2,1)
					health = max_health
					dead = 0
					icon_state = "green"

	Login()
		. = ..()

		power = 20
		health = 50
		max_health = 50

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target
			state

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return
			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target, 2)
					sleep(10)
			else
				step_rand(src)
				sleep(20)

			spawn() ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()

			call_for_help()

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return
				if(target) return
				target = m

		blue
			power = 10
			health = 30
			max_health = 30

			team = TEAM_2
			icon_state = "blue"

			state = NORMAL

			var
				turf/hiding_spot

			// The AI for the mob is very simple. When it gets weak
			// it picks a hiding spot from the global list and
			// transitions to the HIDE state. When in the HIDE state
			// the mob runs towards its selected hiding spot.
			ai()
				if(dead) return

				if(state == NORMAL)
					get_target()

					if(health < 10)
						world << "[name] runs and hides!"

						// hiding_spots is a global list of the locations of
						// all instances of /obj/hiding_spot.
						hiding_spot = pick(hiding_spots)
						state = HIDE
						return ai()

					if(target)
						if(get_dist(src, target) <= 1)
							attack(target, 1)
							sleep(3)
						else
							step_to(src, target)
							sleep(4)
					else
						step_rand(src)
						sleep(20)

				else if(state == HIDE)
					if(!step_to(src, hiding_spot))
						step_towards(src, hiding_spot)
					sleep(2)

				spawn() ai()

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// we use a very simple get_target proc
			get_target()
				for(var/mob/m in oview(3, src))
					if(!m.dead && m.team != team)
						target = m
						return

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m