/*

In all of the examples the ai proc has always had long
delays. In an actual game you'd want mobs to be more
responsive (so they react faster) so you'd lower these
delays. For analyzing and understanding the behavior it
can help to have longer delays.

I've ignored this topic until now so that the ai proc
was as simple as possible. This way the code for each new
feature would not be lost in a mess of unrelated code. In
this example we'll look at what changes need to be made
to handle AI cycles with shorter delays.


Here's the ai proc from code-06.dm:

		ai()
			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
					sleep(10)
			else
				step_rand(src)
				sleep(20)

			spawn() ai()

There are three delays, one after each type of action the mob
can take. If the mob is chasing after you it only has a delay
of 5 (half a second). If the mob attacks you it has a delay of
10 (one second). The reason this delay is longer is because I
wanted to limit the enemy's attack rate to one attack per second.
The problem is that all actions are suspended for that second,
not just attacks.

When the mob calls step_rand there is a two second delay before
the mob plans its next move. In these two seconds many things
could happen, you could run right past the mob! By the time it
starts to plan its next action, you could already be past it.

Here's one way we can address this problem:

		var
			attack_delay = 0
			move_delay = 0
			wander_delay = 0
		ai()
			attack_delay -= 1
			move_delay -= 1
			wander_delay -= 1

			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					if(move_delay <= 0)
						step_to(src, target)
						move_delay = 2
				else
					if(attack_delay <= 0)
						attack(target)
						attack_delay = 5
			else
				if(wander_delay <= 0)
					step_rand(src)
					wander_delay = 10

			spawn(2) ai()

The only delay between AI cycles is introduced by the spawn(2)
in the last line of the proc. The delay variables we added are
used to determine if an action can be performed. Every cycle
subtracts one from these counters. You can perform the action
only if the appropriate counter is <= 0. When you perform the
action the delay counter is set.

With this code, when step_rand is called there is a two second
delay before it can call step_rand again. However, the ai proc
will tick every 0.2 seconds until that happens. During one of
those ticks the mob might acquire a target and start chasing
after it.

This example is a modified version of code-14.dm. The changes were
made in the ai procs for /mob/ai and /mob/ai/blue.


  -- When you run this example --

Compare this example to code-14.dm. Run one, then the other.
In both examples run up to a mob and see how long it takes
for the mob to attack you. In this example it should never
take more than 0.2 seconds for the mob to realize you're there.

In the example from code-14.dm you might time it just right and
approach a mob just as its ai cycle begins (and it attacks you
right away). Other times you might approach the mob right after
its ai cycle ends and it takes almost two seconds before the
mob seems to notice you.

*/


var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		NORMAL = 1
		HEAL = 2
		ATTACK = 3

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src, 3)

	Stat()
		stat(src)
		stat("team", team)
		stat("health", health)

		for(var/mob/ai/m in world)
			stat(" ")
			stat("AI mob", m)
			stat("team", m.team)
			stat("health", m.health)
			stat("target", m.target)

mob

	var
		team = 0
		power = 0
		health = 0
		max_health = 0

	proc
		ai()
		hit_by(mob/m)
		healed_by(mob/m)

		attack(mob/target, damage)
			if(target.dead) return

			show_hit(target)
			target.hit_by(src)

			target.health -= damage
			if(target.health <= 0)
				target.die()
				killed(target)

		heal(mob/target, healing)
			if(target.dead) return

			show_heal(target)
			target.healed_by(src)

			target.health += healing
			if(target.health > target.max_health)
				target.health = target.max_health

		killed(mob/m)
		die()
			if(!dead)
				world << "[name] dies."
				dead = 1
				icon_state = "[icon_state] dead"

				if(client)
					client << "You will resurrect in 3 seconds..."
					sleep(30)
					loc = locate(2,2,1)
					health = max_health
					dead = 0
					icon_state = "green"

	Login()
		. = ..()

		power = 20
		health = 50
		max_health = 50

		team = TEAM_1
		icon_state = "green"

	ai
		var
			mob/target
			state

			attack_delay = 0
			move_delay = 0
			wander_delay = 0

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			if(dead) return

			attack_delay -= 1
			move_delay -= 1
			wander_delay -= 1

			get_target()

			if(target)
				call_for_help()

				if(get_dist(src, target) > 1)
					if(move_delay <= 0)
						step_to(src, target)
						move_delay = 2
				else
					if(attack_delay <= 0)
						attack(target)
						attack_delay = 5
			else
				if(wander_delay <= 0)
					step_rand(src)
					wander_delay = 10

			spawn(2) ai()

		hit_by(mob/m)
			change_target(m)

		proc
			get_target()

			call_for_help()

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return
				if(target) return
				target = m

		red
			power = 30
			health = 40
			max_health = 40
			team = TEAM_2
			icon_state = "red"

			get_target()
				for(var/mob/m in oview(2, src))
					change_target(m)

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.power > target.power)
					target = m

		blue
			power = 10
			health = 30
			max_health = 30

			team = TEAM_2
			icon_state = "blue"

			state = NORMAL

			ai()
				if(dead) return

				// We only worry about the delay for wandering because it was
				// the most severe (two seconds). We still call step_rand every
				// two seconds because of the delay counter, but ai() gets called
				// many times in between calls to step_rand so that we can try
				// to get a target before the full two second delay is up.
				wander_delay -= 1

				if(state == NORMAL)
					get_target()

					// If we get a target we transition to the ATTACK or HEAL state.
					if(target)
						// If we transition to a new state we call the ai proc
						// over again and return (which aborts this iteration
						// of the ai cycle). The reason we do this is because
						// starting a new cycle will cause us to execute the
						// proper action for the state that we just transitioned
						// to.
						// The problem is that calling ai() again this quickly
						// can mess up our delay counters by decrementing them
						// too quickly. In this case, since we're transitioning
						// to HEAL or ATTACK (both states were we won't call
						// step_rand) we probably won't affect the delay between
						// random steps.
						if(target.team == team)
							state = HEAL
							sleep(1)
							return ai()
						else
							state = ATTACK
							sleep(1)
							return ai()
					else
						if(wander_delay <= 0)
							step_rand(src)
							wander_delay = 10
						sleep(2)

				else if(state == ATTACK)
					// If we've lost our target, go back to the NORMAL state
					if(!target || target.dead)
						state = NORMAL
						sleep(1)
						return ai()

					// If we're within range, make the attack and transition
					// back to the NORMAL state.
					if(get_dist(src, target) <= 3)
						attack(target, 2)
						sleep(10)
						state = NORMAL

					// otherwise move closer.
					else
						step_to(src, target)
						sleep(5)

				// Healing is very similar to attacking. If you're not within
				// range you need to move closer. Once you're within range,
				// perform the heal and go back to the NORMAL state.
				else if(state == HEAL)
					// If we've lost our target, go back to the NORMAL state
					if(!target || target.dead)
						state = NORMAL
						sleep(1)
						return ai()

					if(get_dist(src, target) <= 5)
						heal(target, 10)
						sleep(10)
						state = NORMAL
					else
						step_to(src, target)
						sleep(5)

				spawn() ai()

			call_for_help()
				for(var/mob/ai/m in oview(5, src))
					if(m.team == team)
						m.change_target(target)

			// The only changes made to this function were to
			// increase the range that the mob considers. Even though
			// the blue mob can only heal allies within 5 tiles, it
			// looks within 8 tiles to find a weak ally. A similar
			// change was made for finding an enemy to attack.
			get_target()
				target = null
				var/most_utility = 0

				for(var/mob/m in view(8, src))
					if(!m.dead && m.team == team)
						if(m.health == m.max_health) continue
						var/utility = 5 / (m.health)
						if(utility > most_utility)
							most_utility = utility
							target = m

				for(var/mob/m in oview(8, src))
					if(!m.dead && m.team != team)
						var/utility = 10 / (m.health + 10)
						if(utility > most_utility)
							most_utility = utility
							target = m

			change_target(mob/m)
				if(!m) return
				if(m.dead) return
				if(m.team == team) return

				if(!target)
					target = m
					return

				if(m.health < target.health)
					target = m