/*

This example shows how to make mobs that chase and track down
the player without cheating.

It would be very easy to create mobs that chase you. Every AI
cycle you'd just call step_to(src, target). The mob will continue
to chase you and easily be able to find you even if you run
out of their view.

In this example mobs will step_to you if you're in their view.
If you're not in their view they'll step_to the last location
they saw you at (in hopes they when they reach there you'll be
in their view again). This is similar to how a human player
would chase after you.

It would be easier to implement the first method that I mentioned,
but sometimes more human-like results are preferred.

This example was built off of code-03.dm. Because we're only dealing
with mobs that chase we don't need all of the code for attacking
that the later examples have.


  -- When you run this example --

Get the red mob to chase you, then try to lose it. If you can't see
the red mob from your current position it can't see you. Once it
gets out of your sight, stop and see if it catches up to you.

Every time the mob isn't able to catch up to you after you run
out of its sight, that's not a failure of the AI that's a triumph
of its realism! The simplest implementation would always call
step_to(src, target) and the red mob would always catch you, no
matter how tricky your evasive maneuvers are.


  -- What to learn from this --

Sometimes we want AI-controlled players to behave like human players
and sometimes we don't.	It's often easier to create AI players that
don't behave like humans but this can lead to boring or unfair gameplay.

There's no single trick to making AI-controlled players act more like
humans, it's a large bag of tricks:

In code-14.dm we used a state machine and utility functions to make
the blue mobs intelligently use their healing and attack abilities,
similar to how a human player might: only heal allies when they get
weak, otherwise attack the target. The mobs are also capable of moving
to a proper position in order to execute each action.

In code-15.dm we placed invisible markers on the map to indicate
locations that mobs should run towards to hide when they get weak.
This might seem like non-human behavior since humans cannot utilize
these invisible markers. If we place these markers in locations where
human players might hide we create human-like behavior.

In this example we modeled imperfect tracking methods to simulate how
a human player would chase after someone.

These three examples are all very different but their goals were the same:
to create AI-controlled players that behave like human-controlled players.

*/

// We want to have opaque walls for this example.
turf
	wall
		opacity = 1

mob
	proc
		ai()

	Login()
		. = ..()
		icon_state = "green"

		// For this example we don't need the blue mobs,
		// but I don't want to make a separate map without
		// them so we'll just delete them at runtime.
		for(var/mob/ai/blue/b in world)
			del b

	ai
		var
			mob/target

			// You'll see how this new variable is used in the ai proc.
			turf/last_known_loc = null

		New()
			..()
			spawn(rand(10,20)) ai()

		ai()
			get_target()

			if(target)
				// if the target is in our view, move towards it and
				// make a note of its location.
				if(target in view(10,src))
					step_to(src, target)
					last_known_loc = target.loc
					sleep(4)

				// otherwise the target is not in our sight, so we should
				// move towards its last known location.
				else
					if(last_known_loc)
						// If we're not at the last known location,
						// continue moving towards it.
						if(loc != last_known_loc)
							step_to(src, last_known_loc)
							sleep(4)

						// Once we reach our target's last known location
						// if we still can't see them we should go back to
						// random wandering
						else
							step_rand(src)
							target = null
							sleep(4)

					// If we don't have a last known location for the
					// target, go back to wandering randomly.
					else
						target = null
						sleep(1)
			else
				step_rand(src)
				sleep(5)

			spawn() ai()

		proc
			get_target()
				for(var/mob/m in oview(10,src))
					target = m
					return

		red
			icon_state = "red"
		blue
			icon_state = "blue"
