/*

 ** IMPORTANT NOTE ** use map-3.dmi for this example.

Make sure that map-3.dmm is the only selected map. Both map-1.dmm and
map-2.dmm should be unchecked.


The problem in code-19a.dm is that we have 3000 enemies. Most of them can't
be seen by the player, but they're still executing their AI loops. Do they
need to? If the enemy sits idle but the player can't see that, does it
matter?

That's the basis of our fix. When no players are nearby the enemies will
sleep (their AI loop will stop). When a player is nearby the enemies will
be woken up.

One way to do this is to have the enemies repeatedly check if a player is
nearby. And, if one is, the AI mob wakes up. The problem is that this is
almost as bad as running the 3000 AI loops.

Instead we wake up nearby enemies every time the player moves. Even if you
have 100 players in your game, having 100 players checking for sleeping
enemies is better than having 3000 enemies checking for nearby players.

We wake mobs within a large range of the player, larger than their view. This
way the mobs appear active when they do come into view. The problem is, we
might wake up mobs that never come into view. Wandering mobs (ones that aren't
chasing after their target) will check for nearby players. If no players are
nearby they'll go back to sleep.


  -- When you run this example --

The CPU usage should be zero. Initially there are no enemies near you so nothing
should be happening. Even if you find a large group of mobs and wake up 50 of them,
that's a lot better (for the CPU to handle) than having all 3000 AI loops running
at the same time.

The stat panel now displays the number of sleeping mobs and the number of awake
mobs. Run towards enemies and see that the awake count goes up. Run away to see
that they go back to sleep.

Run around and see how many you can wake up, then run away to see if they all go
back to sleep.

*/


var
	const
		TEAM_1 = 1
		TEAM_2 = 2
		TEAM_3 = 3

		ENEMY_COUNT = 3000

	sleeping_mobs = ENEMY_COUNT

world
	view = 10
	mob = /mob/player
	New()
		for(var/i = 1 to ENEMY_COUNT)
			while(1)
				var/x = rand(1,200)
				var/y = rand(1,200)

				if(x < 20 && y < 20) continue

				var/turf/t = locate(x,y,1)
				if(t.density) continue

				if(prob(50))
					new /mob/ai/blue(t)
				else
					new /mob/ai/red(t)
				break

turf
	wall
		opacity = 1

mob
	Click()
		if(get_dist(src, usr) <= 1)
			usr.attack(src)

	Stat()
		stat("CPU", world.cpu)
		stat("# sleeping", sleeping_mobs)
		stat("# awake", ENEMY_COUNT - sleeping_mobs)

mob
	var
		team = 0
	proc
		ai()
		hit_by(mob/m)
		attack(mob/target)
			show_hit(target)
			target.hit_by(src)

	player
		Login()
			. = ..()
			team = TEAM_1
			icon_state = "green"

		Move()
			. = ..()
			// The player's view size is 10, so if we wake all mobs within
			// 13 tiles we'll wake some mobs just outside our range of view.
			// This way, when they do come into view they're already active.
			for(var/mob/ai/m in range(13,src))
				if(m.sleeping)
					m.sleeping = 0
					sleeping_mobs -= 1
					spawn() m.ai()

	ai
		var
			mob/target

			// Initially all AI mobs are sleeping
			sleeping = 1

		// We don't want the AI loop to start when the mob is created,
		// we want it to start when they are woken up.
		/*
		New()
			..()
			spawn(rand(10,20)) ai()
		*/

		ai()
			if(target)
				if(get_dist(src, target) > 1)
					step_to(src, target)
					sleep(5)
				else
					attack(target)
					sleep(10)
			else
				get_target()
				if(target)
					sleep(5)
				else

					// If no players are nearby, go back to sleep.
					var/mob/player/p = locate(/mob/player) in range(16, src)
					if(!p)
						sleeping_mobs += 1
						sleeping = 1
						return

					step_rand(src)
					sleep(20)

			spawn() ai()

		hit_by(mob/m)
			if(m.team == team) return

			if(!target)
				target = m

		proc
			get_target()

		red
			team = TEAM_2
			icon_state = "red"

			get_target()
				for(var/mob/m in oview(2, src))
					if(m.team != team)
						target = m
						return

		blue
			team = TEAM_2
			icon_state = "blue"